# -*- coding: cp1252 -*-

'''
Gestion des diffrents automates:
    M3Zelio, mthodes publics:
        __init__() : verifie les paramtres et initialise l'objet
        open() : ouvre le port
        close() : ferme le port
        read() : lit les donnes
        write() : crit les donnes
        les autres mthodes et variables sont prives (forme __xxx)
    Autre:
'''

import  serial      as      ser
import  logging     as      log
import  time
from    defGen      import  abort

# Gestion des automates M3 et Zlio
class M3Zelio(object):

    # encadrement trame
    __FRAME_BEGIN   = ':'
    __FRAME_END     = '\r\n'

    # taille des donnes traites
    __DATA_FMT      = '%02X'                                    # format des donnes
    __CRC_LNG       = 1                                         # taille CRC (en octets)
    
    __ADDR_OFFSET   = -1                                        # offset des adresses automate
    __ADDR_HEAD     = [0, 0, 0x0ff]                             # entte adresse automate
    
    # Init objet, les variables __slaveAddr et __CRCAddr sont prives
    def __init__(self, Machine, port, readAddr=1, readCount=0, writeAddr=1, writeCount=0):

        self.port = port
        self.readAddr = readAddr
        self.writeAddr = writeAddr

        # Vrifie le type d'automate, initialise __slaveAddr et __CRCAddr
        machine = Machine.lower()
        machines = {'m3':[4, 1], 'zelio':[1, 2]}    # Dictionnaire des automates grs, format = machine:[__slaveAddr, __CRCAddr]
        if machine not in machines.keys():
            abort('Type d\'automate invalide : %s , doit tre %s (en majuscule ou minuscule)' %
                  (Machine, ' ou '.join('%s' % machine for machine in machines.keys())))
        self.__slaveAddr = machines[machine][0]
        self.__CRCAddr = machines[machine][1]

        # Vrifie les paramtres de lecture de l'automate
        self.__controllerDataVerify('lire', readAddr, readCount, 1, 48)

        # Vrifie les paramtres d'criture de l'automate
        self.__controllerDataVerify('crire', writeAddr, writeCount, 1, 24)

    # Vrifie les paramtres de l'automate
    def __controllerDataVerify(self, msg, addr, count, firstAddr, lastAddr):

        # Vrifie l'adresse de la donne
        if addr < firstAddr or addr > lastAddr:
            abort('Adresse 1re donne automate  %s invalide : %s, doit tre comprise entre %s et %s' % (msg, addr, firstAddr, lastAddr))

        # Vrifie le nombre de donnes
        maxCount = lastAddr - firstAddr + 1
        if count < 0 or count > maxCount:
            abort('Nombre de donnes automate  %s invalide : %s, doit tre compris entre 0 et %s' % (msg, count, maxCount))

        # Vrifie la cohrence entre adresse donne et nombre de donnes
        if addr + count - 1 > lastAddr:                         # suivant les spcifications
            abort('adresse dernire donne automate  %s invalide : %s,\n'
                  'la somme (1re adresse + nombre de donnes - 1) doit tre infrieure  %s' % (msg, addr+count-1, lastAddr+1))

    # ouverture port, la variable __port est prive
    def open(self):
        self.__port = ser.Serial(self.port, baudrate=115200, bytesize=ser.SEVENBITS, parity=ser.PARITY_EVEN, stopbits=ser.STOPBITS_ONE, timeout=10, writeTimeout=10)
        return self.__port

    # fermeture port
    def close(self):
        self.__port.close()

    def read(self, dataCount16=0, readAddr=None):
        '''
        Demande des donnes SLin/SLout  l'automate et renvoie sa rponse
        Noter que les SLout commencent  25 et sont les seuls dans lesquels le programme automate peut crire.
            Entres:
                dataCount16: nombre de donnes 16 bits signes  lire (de 0  48)
                readAddr: optionnel, adresse de lecture dans l'automate
            Retour: liste de 0  48 entiers 16 bits signs, None si erreur
        '''

        READ_CMD = 0x03

        if readAddr is None:
            readAddr = self.readAddr                            # adresse de lecture initialise avec l'objet M3Zelio

        # prpare le message de requte pour l'automate
        dataCount8 = dataCount16 * 2                            # nombre de donnes  lire (fois 2 car sur 16 bits)
        # entte requte: adresse esclave, commande lecture, adresse donnes (__ADDR_HEAD, adresse de lecture+__ADDR_OFFSET), nombre de donnes  lire
        reqHead = [self.__slaveAddr, READ_CMD]; reqHead.extend(self.__ADDR_HEAD); reqHead.extend([readAddr+self.__ADDR_OFFSET, dataCount8])
        msg = self.__FRAME_BEGIN + ''.join(self.__DATA_FMT % val for val in reqHead) + \
              ''.join(self.__DATA_FMT % val for val in self.__CRCList(reqHead)) + self.__FRAME_END

        # envoie la requte et gre la rponse de l'automate
        list8 = self.__getAnswer(self.__port, msg, 'lecture', [self.__slaveAddr, READ_CMD, dataCount8])

        if list8 is None:                                       # erreur dans la rponse automate
            return None

        # vrifie que le nombre de donnes 8 bits reues est correct
        if len(list8) != dataCount8:
            log.info(u'Erreur lecture automate, nombre de donnes 8 bits lues invalide: %s' % list8)
            return None

        # construit des entiers 16 bits signs  partir des donnes 8 bits lues (s'il y en a)
        list16 = []
        for idx in range(0, len(list8), 2):
            data16 = list8[idx] << 8 | list8[idx+1]             # poids fort et faible
            if list8[idx] & 0x080:
                data16 -= 0x010000                              # nombre ngatif
            list16.append(data16)

        return list16                                           # retourne la liste des donnes 16 bits signes

    def write(self, list16=[], writeAddr=None):
        '''Ecrit des donnes SLin dans l'automate et renvoie sa rponse
        Les donnes modifies doivent tre conscutives.
            Entres :
                list16: liste de valeurs 16 bits signes (liste vide ou une seule valeur)  crire dans l'automate, de 0  24 lments
                writeAddr: optionnel, adresse d'criture dans l'automate
            Retour : [] si OK, None si erreur
        '''

        WRITE_CMD = 0x10

        # si 1 seul lment fourni, le transforme en itrable
        if not hasattr(list16, '__iter__'):
            list16 = [list16]

        if writeAddr is None:
            writeAddr = self.writeAddr                          # adresse d'criture initialise avec l'objet M3Zelio

        # prpare le message de requte pour l' automate
        # entte requte: adresse esclave, commande criture, adresse donnes (__ADDR_HEAD, adresse d'criture+__ADDR_OFFSET), nombre de donnes  crire (fois 2 car sur 16 bits)
        reqHead = [self.__slaveAddr, WRITE_CMD]; reqHead.extend(self.__ADDR_HEAD); reqHead.extend([writeAddr+self.__ADDR_OFFSET, len(list16)*2])
        CRC = 0
        data = ''
        for val in list16:                                      # ajout des donnes 16 bits (en 2 fois 8 bits)
            hval = (val >> 8) & 0x0ff                           # poids fort
            lval = val & 0x0ff                                  # poids faible
            data += self.__DATA_FMT*2 % (hval, lval)            # ajout poids fort et faible
            CRC += hval + lval                                  # mise  jour CRC
        msg = self.__FRAME_BEGIN + ''.join(self.__DATA_FMT % val for val in reqHead) + data + \
              ''.join(self.__DATA_FMT % val for val in self.__CRCList(reqHead, CRC)) + self.__FRAME_END
        
        # envoie la requte et gre la rponse de l'automate (None en cas d'erreur)
        return self.__getAnswer(self.__port, msg, 'criture', reqHead)
        
    # tronc commun des mthodes read et write
    def __getAnswer(self, port, reqMsg, fonction, repHead):
        ''' Mthode prive de la classe M3Zelio, envoie la requte  l'automate, attend sa rponse,
            la vrifie (encadrement, caractres hexa, longueur, CRC) et retourne les donnes utiles.
            Entres :
                port : port srie de l'automate (doit tre ouvert)
                reqMsg : message de requte
                fonction : 'lecture' ou 'criture'
                repHead : entte attendue dans la rponde automate
            Retour : liste d'entiers 8 bits en lecture, liste vide en criture, None si erreur.
        format de la trame rponse automate: __FRAME_BEGIN + suite de caractres hexa (par groupe de 2) + __FRAME_END
        '''

        repHeadLng = len(repHead)                               # longueur de l'entte attendue en rponse
        
        # vide le tampon d'entre et envoie la requte  l'automate
        port.flushOutput()
        port.flushInput()
        port.write(reqMsg)
    
        # attend le 1er caractre de l'automate ou le timeout de lecture
        s = port.read(1)

        # teste si on a eu 1 caractre    
        if not s:
        
            # on a rien reu dans les temps
            log.info(u'Erreur, pas de rponse de l\'automate')
            return None

        # on attend encore 1 seconde pour que tout le message soit l
        # on a le 1er caractre, on lit le reste du message et on met
        #  0 le bit 7 de tous les caractres (bit de parit)
        time.sleep(1)
        msg = ''.join(chr(ord(c) & 0x07f) for c in s+port.read(port.inWaiting()))

        # verifie l'encadrement de la trame reue
        if msg[:len(self.__FRAME_BEGIN)] != self.__FRAME_BEGIN or msg[-len(self.__FRAME_END):] != self.__FRAME_END:
            log.info(u'Erreur, message automate avec caractres d\'encadrement invalides: %s' % msg)
            return None

        # supprime l'encadrement
        msg = msg[len(self.__FRAME_BEGIN):-len(self.__FRAME_END)]

        # vrifie que le nombre de caractres restant est pair (2 caractres hexa donne un entier 8 bits)
        if len(msg) % 2:
            log.info(u'Erreur, message automate comportant un nombre impair de caractres: %s' % msg)
            return None

        # vrifie que tous les caractres sont des chiffres hexa et les convertit en entiers (par groupe de 2 caractres)
        try:
            liste = [int(msg[idx:idx+2], 16) for idx in range(0, len(msg), 2)]
        except ValueError, TypeError:
            log.info(u'Erreur, message automate avec caractres non hexa')        
            return None

        # vrifie si la partie utile du message automate a au moins la taille du header attendu
        if len(liste) - self.__CRC_LNG < repHeadLng:            # -__CRC_LNG pour ne pas tenir compte du CRC
            log.info(u'Erreur, message automate trop court (%s caractres min), message reu: %s' % (repHeadLng, liste[:-1]))
            return None

        # vrifie si le header reu est conforme
        if liste[:repHeadLng] != repHead:
            log.info(u'Erreur %s automate, entte message invalide: %s, attendu: %s' % (fonction, liste[:repHeadLng], repHead))
            return None

        # calcule et vrifie le CRC
        CRC = self.__CRCCalc(liste[:-self.__CRC_LNG])           # calcul CRC  partir des donnes trame
        if CRC != self.__CRCVal(liste):                         # diffrent du CRC indiqu en fin de trame ?
            log.info(u'Erreur, message automate avec CRC incorrect')
            return None

        return liste[repHeadLng:-self.__CRC_LNG]                # retourne la liste des donnes (s'il y en a)

    # Calcul CRC
    def __CRCCalc(self, liste):
        return (~sum(liste) + self.__CRCAddr) & ((1 << (8 * self.__CRC_LNG)) - 1)

    # Calcul CRC sous forme de liste
    def __CRCList(self, liste, initVal=0):
        liste = list(liste)                                     # pour conserver la liste originale...
        if initVal:
            liste.append(initVal)                               # ajout d'une valeur de dpart
        CRC = self.__CRCCalc(liste)                             # calcul CRC
        List = []                                               # constitue la liste
        for idx in range(self.__CRC_LNG):
            poids = self.__poids(idx)
            List.append(CRC // poids)
            CRC %= poids
        return List
        
    # Calcul du CRC indiqu en fin de liste
    def __CRCVal(self, liste):
        somme = 0
        for idx, val in enumerate(liste[-self.__CRC_LNG:]):
            somme += val * self.__poids(idx)
        return somme

    # poids (1, 256, 65536...)
    def __poids(self, idx):
        return 1 << (8 * (self.__CRC_LNG - idx -1))

